/*++++++++++++++
.IDENTIFICATION tabout.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    
.KEYWORDS       ASTRORES
.VERSION  1.0   29-Apr-2001
.VERSION  1.1   02-Jun-2007: Cosmetics (-Wall)
.COMMENTS       Output from a TABLE a row 
The basic "language" to generate a link is
${colname}	: replace by the contents of the column
${+colname}	: replace by the contents of the column, HTTP-encoded
${*colname}	: replace by the TRIMMED contents of the column
${**colname}	: replace by the TAIL TRIM contents of the column
	${}	= contents of current column
	${+}    = contents of current column, HTTP-encoded
	${*}	= contents of trimmed current column
	${**}	= contents of TAIL trimmed current column
	${***}	= contents of column, ALL BLANKS removed
	${%d}	= contents of (decimal edited) current column
	${$cat} = name of catalog
	${$tab} = name of table
	${$col} = name of column
	${$catid} 	= Number of Catalog
	${$pos} 	= position defined in this row
	${$poseq} 	= position and equinox, e.g. '12:23.3+12:23,eq=J2000'
	${name}		= contents of column 'name'
	${*name}	= contents of column 'name', trimmed
	${=name}	= argument for column 'name': htarg{name=value}
	${=*name}	= argument for column 'name', trimmed
	${=**name}	= argument for column 'name', TAIL trimmed
	${==name}	= argument for column 'name': htarg{name=value}
	${==*name}	= argument for column 'name', trimmed
	${==**name}	= argument for column 'name', TAIL trimmed
	${name$1-n*}	= contents of column 'name', chars 1-n, see (*)
	${name$c1-n*}	= contents of column 'name', fields 1-n, separator c
	${name$(n)*}	= contents of column 'name', nth parenthesis content
	${name$<n>*}	= contents of column 'name', nth angle bracket content
	${name$[n]*}	= contents of column 'name', nth square bracket content
	(*)		= Terminating Character, may be
			z = transform blanks into zeroes
			b = blank leading zeroes
			d = delete specified chars
We assume the macro
---------------*/

#include <catstd.h>
#include <stdio.h>		
#include <stdlib.h>		
#include <string.h>		
#include <buffer.h>
void *malloc() ;

static int  rowlen ;
static char *colname ;
static char brackets[] = "()[]{}<>";

/*===========================================================================
		Internal Utilities
 *===========================================================================*/

static int subfields(char *buf, char *string, char *comp)
/*++++++++++++++++
.PURPOSE  Concatenate subfields, e.g 1-3,6,... /1,2
.RETURNS  The length of filled buf
-----------------*/
{
  static char known_conv[] = "bzd" ;
  char *sbuf ;
  char sep, conv, sep1;
  int  i1, i2, i ;
  char *p, *s ;

    sbuf = buf;

    sep = sep1 = conv = 0 ;
    if (!*comp) {		/* No component specified => copy all */
	strcpy(sbuf, string);
	sbuf += strlen(sbuf);
    }

    while (*comp && (!conv)) {
    	if (!isdigit(*comp)) { 
	    if (strchr(known_conv, *comp)) conv = *comp ;
	    else sep = *comp ;
	    comp++ ; 
	    continue ; 
	}
	i1 = i2 = atoi(comp) ; while (isdigit(*comp)) comp++ ;
	if ((*comp == '-') || (*comp == '/')) {
	    comp++ ;
	    i2 = atoi(comp) ; while (isdigit(*comp)) comp++ ;
	    if (i2 == 0) i2 = strlen(string) ;
	}
	else if (*comp == ',') comp++ ;
	if (i1 > 0) i1-- ;
	i2 -= i1 ;		/* Number of items */
	if (sep) {
	    /* Accept $(1) = 1st parenthesis, $[2] = 2nd bracket, etc */
	    i = strloc(brackets, sep);
	    if (brackets[i] && (brackets[i+1] == *comp)) {
		sep1 = *(comp++);
		i1++; 		/* (1) is 2nd word with separator ( ) */
	    }
	    for (p=string ; *p ; p++)  { 
	        if (i1 <= 0) break ;
	        if (*p == sep) i1-- ; 
	    }
	    if (!i2) i2 = 1;	/* At least 1 subfield */
	}
	else for (p=string ; *p && (--i1 >= 0); p++) ;
	while ((*p) && (--i2 >= 0)) {
	    if (sep)  {
		if (sep1) i = strloc(p, sep1);
		else      i = strloc(p, sep) ;
		if (p[i] && (i2 > 0)) i++ ;
	    }
	    else i = 1 ;
	    strncpy(sbuf, p, i), sbuf += i;
	    p += i ;
	}
	if (conv) break ;
    }
    *sbuf = 0;

    /* Convert if necessary */
    switch(conv) {
      case 0:
	break ;
      case 'b':	/* Blank out leading zeroes */
	for (p=buf; *p == '0'; p++) *p = ' ' ;
	if (!*p) p[-1] = '0' ;
	break ;
      case 'z':	/* Zero leading blanks */
	for (p=buf; *p == ' '; p++) *p = '0' ;
	break ;
      case 'd': /* Delete specified chars */
	for (s=p=buf; *s; s++) { 	/* Delete chars in comp */
	    if (strchr(comp, *s)) continue ;
	    *(p++) = *s ;
	}
	*p = 0 ;
	break ;
      default:
    	fprintf(stderr, "++++Unknown conversion subfield($...%c)\n", conv) ;
	break ;
    }
    return(sbuf-buf) ;
}

static char *value(FILINFO *table, char *name, char *row)
/*++++++++++++++++
.PURPOSE  Find out the value of a column
.RETURNS  The value / NULL when field is BLANK
.REMARKS  Set up the static   colname reclen
-----------------*/
{
  static char *record ;
  static int reclen ;
  COLINFO *pc;
  TABINFO *pt; 
  char *psub, *b;
  int c, w ;

    if (rowlen == 0) {
	rowlen = strlen(row) ;
	if (reclen <= table->reclen) {
	    if (reclen) free(record) ;
	    reclen = ((reclen+1)|63) + 1 ;
	    record = malloc(reclen) ;
	}
    }
    colname = "" ;		/* Yet unknown	*/
    pt = table->tabform ;

#if 0				/* Not Yet Accessible -- will need ${$pos} */
    buf[0] = 0 ;
    if (name[0] == '$') { 	/* Special 	*/
	table = col->table ;
	switch(name[1]) {
	  case 'p':		/* Position */
	    if (strncmp(name, "$pos", 4) == 0) {
		mainpos[0] = mainpos[1] = (COLINFO *)0 ;
		for (ac=table->acol, ec=ac+table->ncol; ac<ec; ac++){
		    if (!((*ac)->otable&VO_MAINPOS)) continue ;
		    if ((*ac)->name[0] == 'p') continue ;	/* Proper Mo. */
		    mainpos[(*ac)->type&VM_PERIODIC ? 0 : 1 ] = *ac ;
		}
		if ((!mainpos[0]) || (!mainpos[1])) {
		    fprintf(stderr, "****No position in table '%s'\n",
		      table->name );
		    return(buf) ;
		}
		if (!(mainpos[0]->otable & VO_EDITED)) meta_edit(mainpos[0]) ;
		if (!(mainpos[1]->otable & VO_EDITED)) meta_edit(mainpos[1]) ;
		if (isNaN(&(mainpos[0]->value.f))) return((char *)0) ;
		if (isNaN(&(mainpos[1]->value.f))) return((char *)0) ;
		else sprintf(buf, "%s%s", 
		    mainpos[0]->avalue, mainpos[1]->avalue) ;
	        if (strncmp(name, "$poseq", 6) == 0) {	/* Add Equinox */
		    f = buf + strlen(buf) ;
		    f[0] = ','; f[1] = 'e'; f[2] = 'q'; f[3] = '=' ;
		    if (mainpos[0]->name[0] == 'R') {	/* Equatorial */
			if (table->equinox == 2000.) *f = 0 ;
			else sprintf(f+4, "%.0f", table->equinox) ;
		    }
		    else f[4] = mainpos[0]->name[0], f[5] = 0 ;
		}
		return(buf) ;
	    }
	    break ;
	  case 't':
	    if (strcmp(name, "$tab") == 0) 
		return(table->name) ;
	}
    }
    if ((name[0] == '%') && ((col->type&7) == 0)) { /* Re-edit integer value */
	sprintf(buf, name, col->value.i) ;
	return(buf) ;
    }
#endif

    /* Find out the column */
    if ((psub = strchr(name, '$'))) *psub = 0 ;
    pc = tbf_findlabel(pt, name) ;
    if (!pc) { 
	fprintf(stderr, "++++Column '%s' not found in table '%s'\n", 
	    name, table->dfile) ;
	colname = name ;
	if (psub) *psub = '$' ;
	return((char *)0);
    }

    if (pc->b0 >= rowlen) {	/* Empty column */
	if (psub) *psub = '$' ;
	return("") ;
    }

    b = row+pc->b0 ; w = pc->len ;
    if (psub) {			/* Take subfields */
	*(psub++) = '$' ;
	c = b[w]; b[w] = 0 ;
	subfields(record, b, psub) ;
	b[w] = c ;
    }
    else strncpy(record, b, w), record[w] = 0 ;

    return(record) ;
}

/*===========================================================================
		External Routines
 *===========================================================================*/

char *tbf_edrow(FILINFO *tab, char *format, char *row)
/*++++++++++++++++
.PURPOSE  Convert one row, from "format" like ${name} (cf vizout.c)
.RETURNS  The (static) result
.REMARKS  The row is supposed to be right-filled with blanks !
-----------------*/
{
  static BUFFER *abuf ;
  int i, x, len;
  int trim;	/* Number of * -- 1 = trim, 2 = right trim only */
  int argeq;	/* Number of = signs, (add argument and = sign) */
  int cvt ;	/* Requires conversion right / left of = sign.. */
  char *m, *s, *e ;

    if (!abuf) abuf = BUF_New(char, 256, 256) ;
    BUF_Clear(abuf) ; BUF_AppendString(abuf, "") ;
    rowlen = 0 ;

    for (m=format; *m; m += len) {
	len = strloc(m, '$') ;
	if ((len == 0) && (m[1] != '{')) /*}*/ len = 1 ;
	if (len) { BUF_AppendItems(abuf, char, m, len); continue; }

	/* Here len=0, m->${...}, edit a Column */
	cvt = 0 ; if (m[2] == '+') m++, cvt = 3 ;	/* Both name + val */
	for (argeq=0; m[2] == '=' ; m++, argeq++) ;
	if (m[2] == '+') m++, cvt |= 1 ;
	for (trim=0; m[2] == '*' ; m++, trim++ ) ;
	m += 2 ;
	i = strloc(m, /* { */ '}') ; 
	x = m[i] ; m[i] = 0 ;
	if (argeq) cvt |= 3 ;		/* Must encode colname! */
	if ((s = value(tab, m, row))) {	/* colname set by value	*/
	    if (cvt&2) {
		/* ---------- HTML Encoding not Done ********** */
		BUF_AppendString(abuf, colname) ;
		BUF_AppendItem(abuf, char, "=") ;
	    }
	    while (argeq > 1) { argeq--, BUF_AppendItem(abuf, char, "=") ; }
	    e = s + strlen(s) ;
	    if (trim == 1) {		/* Head Trim  */
		while ((s<e) && (*s == ' ')) s++ ;
	    }
	    if (trim) {			/* TAIL trim only */
	        while ((s<e) && (e[-1]==' ')) e-- ;	
	    }
	    /* ---------- HTML Encoding not Done ********** */
	    /* if (cvt == 1) BUF_AppendString(abuf, "\\htarg{") ; */
	    if (trim == 3) while (s < e) {	/* Remove ALL blanks */
		if (*s != ' ') BUF_AppendItem(abuf, char, s) ;
		s++ ;
	    }
	    else {
	        BUF_AppendItems(abuf, char, s, e-s) ;
	        s = e ;
	    }
	    /* ---------- HTML Encoding not Done ********** */
	    /* if (cvt&1) BUF_AppendItem(abuf, char, "}") ; */
	}
	if ((m[i] = x)) m++ ;
	m += i ;
    }

    return(abuf->buf) ;

}

