/*++++++++++++++
.IDENTIFICATION trcat.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    Unix
.KEYWORDS       Convert Catalogue Number
.VERSION  1.0   18-Feb-1992
.VERSION  1.1   06-Feb-1993: Added '@' category
.VERSION  1.2   12-Mar-1993: trcat @a ===> @/yyyy/MMM/dda
.VERSION  1.3   03-Jun-1993: AFOEV --- J with -X option
.VERSION  1.4   26-Mar-1994: There was a bug in edx
.VERSION  1.5   10-Oct-1994: Be sure what follows the cat# is
		not lost...
		when one cat# has an =obsolete= file but no .files
.VERSION  1.6   23-Mar-1995: Y equivalent to @
.VERSION  1.7   13-Jul-1995: Allow lowercase but APJ translated to ApJ
.VERSION  1.8   25-Oct-1995: Verify the version
.VERSION  1.9   24-Jan-1996: Verify Mon/Day directory exists
.VERSION  2.0   15-Apr-1996: Use opendir functions
.VERSION  2.1   03-Oct-1997: Remove the trailing /
.VERSION  2.2   27-Feb-1998: Option -d (Move to directory)
.VERSION  2.3   03-Apr-1998: HTML syndrome  "A A" --> "A+A"
.VERSION  2.4   19-Jun-1998: Includes ADC Site
.VERSION  2.5   22-Jul-1998: Option -r
.VERSION  2.6   11-Dec-1998: Added the 'f'; Be sure to check the version !
.VERSION  2.7   13-Jan-1999: Be sure to separate Directory/./File
.VERSION  2.8   01-Jan-2000: Option -1 for 1 catalogue per line, rest = comment
.VERSION  2.9   10-Jul-2001: Accept cat= , accept /./
.VERSION  2.91  01-Mar-2003: ADC = last item in CATroot list
.VERSION  2.92  31-Dec-2004: Pb afoev fixed.
.VERSION  3.0   23-Jan-2005: Add the table level/special names
.VERSION  3.1   11-Mar-2006: -version; errors messages.
.VERSION  3.2   11-May-2006: Increase MAX_LINE to 10k. Use first $HOME/ftp
.VERSION  3.21  12-Jul-2006: Bug
.VERSION  3.22  17-Jan-2007: Debug option
.VERSION  3.23  02-Jun-2007: Cosmetics (-Wall)
.VERSION  3.24  30-Nov-2007: BaltA AcA
.VERSION  3.25  16-Mar-2008: Option -q (quiet, no error message)
.VERSION  3.26  15-Apr-2008: Convert ' ' into '+' in file names
.VERSION  3.3   27-Aug-2008: Read directories to get the case
.VERSION  3.4   27-Feb-2009: Output the directory name,
			but exit 1 if directory does not exist
.VERSION  3.5   01-May-2009: Accept HTML syndromes (% &amp; ... )
.VERSION  3.6   10-Jun-2009: Replace blanks by underscores
.VERSION  3.7   27-Mar-2010: A+A --> add a 'A' for volume>=509
.VERSION  3.8   03-Aug-2010: Pb with input like 2002A&A...386..816B (bibcode)
.VERSION  3.81  03-Oct-2010: Pb J/other/ApJ/vol/page
.VERSION  3.82  18-Jul-2011: Cosmetics (remove warnings)
.VERSION  3.9   04-Feb-2012: Accept merged otions e.g. -vNd
.VERSION  3.91  06-Feb-2012: Bug with Y/... directories
.VERSION  3.92  12-Feb-2013: Ignore PRIVATE catalogs
.COMMENTS       Translate text of standard input.
---------------*/

#define VERSION  "3.92 (12-Feb-2013)"

#include <stdio.h>
#include <ctype.h>
#include <time.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>

#define MAX_LINE	10240	/* Maximal size of lines read */
#define LEN_CWD		256	/* Maximal length of CWD      */
typedef char	*S;
typedef struct {		/* Dynamical buffer */
	int allocated;
	int increment;
	int used;
	char *buffer;
  } AREA;

#define MAX(x,y)        ((x) >= (y) ? (x) : (y))
#define ITEMS(a)        sizeof(a)/sizeof(a[0])
#define starts_by(s,t)  (!strncmp(s,t,sizeof(t)-1))
#define istarts_by(s,t)	(!strncasecmp(s,t,sizeof(t)-1))
#define ChDir(dir) 	if(chdir(dir)) { fprintf(stderr, "#***chdir(%s): ", \
				    dir); perror(""); }
typedef int (*SORT_FCT)(/*void *, void **/) ;

static char verbop   = 0; /* Verbose option 	*/
static char optf = 0 ;	  /* -df or -rf option  */
static char o1 = 0 ;	  /* -1 option  	*/
static char oq = 0 ;	  /* -q option  	*/
static char vtest = 0 ;	  /* First version	*/
static int cd_status ;	  /* Status of chdir	*/

static char *rm[] = {"X", "I", "II", "III", "IV",
		"V", "VI", "VII", "VIII", "IX" };
static char *pgm;
static char *cats;	/* getenv("CATS") */
static char bwd[LEN_CWD];	/* Original Working Directory */
static char awd[LEN_CWD];	/* Actual   Working Directory */
static char catVersion = 0 ;
static char usage[] = "\
Usage: %s [-1] [-q] [-n|-N|-d[f]|-r[f]|-{x|X|o|p} category] [catalog | .]\n\
";

/* List of possible roots, the first byte = style of Directories */
static char *CATroot[] = {	/* Possible roots */
    "/ftp/cats/", 				/* CDS  */
    "/var/www/cds/ftp/pub/cats/",	 	/* CFA  */
    "/vizier/data/cds/ftp/cats/",		/* CADC */
    "/home/cds/ftp/cats/",			/* Cn   */
    "/home/cds/ftp/",				/* RU   */
    "/cds/ftp/",				/* ADAC */
    "/cds/cds/ftp/",				/* ADAC */
    "/ftp/2/",					/* CDS -- 'more' directories */
    "/disk/adc/ftp/pub/adc/archives/",	 	/* ADC  */
    ""
} ;
static int iCATroot ;		/* Index in above */
#define isADC	(iCATroot == ITEMS(CATroot)-1)

static char *special_names[] = {	/* Number of levels */
    "\2bincats/",
    "\2afoev/",
    "\2AFOEV/",
    "\2more/",
    (char *)0
};
static char *htmspec[] = {
   "&&amp;", "<&lt;", ">&gt;", "\"&quot;", "'&apos;", " &nbsp;", 0 };

/*==================================================================
		Internal Routines
 *==================================================================*/
static void error(text1,text2)
/*++++++++++++++++
.PURPOSE  Error exit
.RETURNS  -
-----------------*/
  char *text1, *text2;     /* IN: Text before error */
{
    if (!oq) {
        fprintf(stderr, "*** %s: %s%s\n", pgm, text1, text2);
        fprintf(stderr, usage, pgm);
    }
    exit(1);
}

static void help()
/*++++++++++++++++
.PURPOSE  Display help, then exit
.RETURNS  To OS
-----------------*/
{
static char text[] = "\
   -v: verbose option\n\
   -1: The first word of each input line is a catalog name, comments printed\n\
   -q: Quiet -- don't issue error message\n\
   -d: Returns the absolute directory name of catalog\n\
  -df: Returns the directory AND file, as  directory/./file\n\
   -n: convert to numeric, e.g. 2001\n\
   -N: convert to Roman style, e.g. II/1\n\
       (default: convert to opposite)\n\
   -r: Returns the relative directory name of catalog\n\
  -rf: Returns the relative directory AND file, as  directory/./file\n\
   -x: expand the category to existing, non-obsolete \n\
   -X: expand the category to ALL (including obsolete) \n\
       each line of input contains a catalogue\n\
   -o: list the obsolete catalogues\n\
   -p: list the private catalogues\n\
 catalog: Catalogues to convert (default from stdin)\n\
       .  for catalogue in the current directory\n\
    Y[n]  for catalogue Y/yyyy/mmm/ddn \n\
Returned status is 1 when catalog directory not found, or bad argument\n\
" ;

    fprintf(stdout, usage, pgm);
    fprintf(stdout, "%s", text);
    exit(0);
}

#if 0
static int strbloc (text, c)
/*++++++++++++++++
.PURPOSE  Locate (backwards) specified character
.RETURNS  Index of located char / -1
-----------------*/
  char *text;       /* IN: String to interpret */
  int  c;		/* IN: Char to look for */
{
    char *s;
    for (s = text + strlen(text); --s >= text; ) if (*s == c)    break;
    return(s-text);
}
#endif

static int strloc(text, c)
/*++++++++++++++++
.PURPOSE  Locate specified character
.RETURNS  Index of located char
-----------------*/
  char *text;       /* IN: String to interpret */
  int  c;		/* IN: Char to look for */
{
  char *s;
    for ( s = text; *s; s++)	if (*s == c)	break;
    return(s-text);
}

static int strcopy(dest, sour)
/*++++++++++++++++
.PURPOSE  Copy, returning the length
.RETURNS  Length
-----------------*/
  char *dest; 	/* OUT: Destination */
  char *sour;	/* IN: Source */
{
  char *d, *s;
    for (d=dest, s=sour; *s; s++, d++)	*d = *s;
    *d = 0;
    return(d-dest);
}

static char *basename(name, sub)
/*++++++++++++++++
.PURPOSE  Extract the 'name' from a complete path
.RETURNS  Converted
-----------------*/
  char *name;	/* IN: Full file name */
  int  sub;	/* IN: Number of Subdirectories to take into account */
{
  char *p;
    for (p = name + strlen(name); --p >= name; ) {
	if (*p == '/') { if (--sub < 0) break; }
    }
    return(++p);
}

static void set_catroot()
/*++++++++++++++++
.PURPOSE  Retrieve the root of all catalogs
.RETURNS  ---
.REMARKS  Set the cats global variable, and moves to this directory !
-----------------*/
{
  char *p, *d;

    if (cats) return;

    if (verbop) fprintf(stderr,
	"... %s: Trying to retrieve the root directory of catalogues...\n",
	pgm);
    if (verbop) fprintf(stderr, "    $CATS");
    cats = getenv("CATS");
    if (!cats) {
	if (verbop) fprintf(stderr, " $Croot");
	cats = getenv("Croot") ;
    }
    if (verbop) {
	if (cats) fprintf(stderr, " => %s", cats);
	fprintf(stderr, "\n");
    }
    if (cats) {
	if (verbop) fprintf(stderr, "    %-32s", cats);
	if (chdir(cats)) {	/* V3.21 */
	    if (verbop) perror(" ");
	    cats = (char *)0;
	}
	goto FIN;
    }

    /* (V3.2) Try first $HOME/ftp */
    if ((d = getenv("HOME"))) {
	cats = malloc(strlen(d) + strlen(CATroot[0]) + 1);
	strcpy(cats, d);
	strcat(cats, CATroot[0]);
	if (verbop) fprintf(stderr, "    %-32s", cats);
	if (chdir(cats)) {
	    if (verbop) perror(" ");
	    free(cats);
	    cats = (char *)0;
	}
    }
    /* Try to guess if CATS not defined */
    for (iCATroot=0; (!cats) && CATroot[iCATroot][0]; iCATroot++) {
	if (verbop) fprintf(stderr, "    %-32s", CATroot[iCATroot]);
	if (chdir(CATroot[iCATroot])) {
	    if (verbop) perror(" ");
	    continue ;
	}
	cats = strdup(CATroot[iCATroot]);
	break ;
    }

    /* Try $HOME/cats */
    if ((!cats) && (d = getenv("HOME"))) {	/* Try on ~/cats */
	p = malloc(strlen(d)+8);
	strcpy(p, d);
	strcat(p, "/cats");
	if (verbop) fprintf(stderr, "    %-32s", p);
	if (chdir(p)) {
	    if (verbop) perror(" ");
	    free(p);
	}
	else cats = p;
    }

  FIN:
    if (cats) {
	/* V3.0: Remove the ending slash */
	p = cats + strlen(cats) -1;
	if (*p == '/') *p = 0;
	if (verbop) fprintf(stderr, " => %s\n", cats);
	return;
    }
    error("(set the CATS environment variable as the root catalog directory)",
	  "");
}

/*==================================================================*/
static int filcmp(c1, c2)
/*++++++++++++++++
.PURPOSE  Compare two strings, taking care of numeric values
.RETURNS  Difference
-----------------*/
  char **c1;	/* IN: First file     */
  char **c2;	/* IN: Second file     */
{
  char *p, *q;
  int diff;

    p = *c1; q = *c2;
    if (p == q)	return(0);	/* Strings are obviously identical ! */
    for (diff=0; (diff == 0) && (*p || *q); ) {
	if (isdigit(*p) && isdigit(*q)) {
	    diff = atoi(p) - atoi(q);
	    while (isdigit(*p)) p++;
	    while (isdigit(*q)) q++;
	    continue;
	}
	diff = *(p++) - *(q++);
    }
    return(diff);
}

/*==================================================================
		AREA Manipulation
 *==================================================================*/
static AREA *alloc(size)
/*++++++++++++++++
.PURPOSE  Get All Files with specified template, keeping only directories
.RETURNS  Area filled with names.
-----------------*/
  int size;	/* IN: Buffer Size */
{
  AREA *a;
  char *buffer;
    a = (AREA *)0;
    buffer = malloc(size); if (!buffer) return(a);
    a = (AREA *)malloc(sizeof(AREA));
    a->increment = a->allocated = size;
    a->used = 0;
    a->buffer = buffer;
    return(a);
}

static int afree(a)
/*++++++++++++++++
.PURPOSE  Append a string in buffer
.RETURNS  0
-----------------*/
  AREA *a;	/* IN: Area to close  */
{
    if (a) {
        if (a->buffer) free(a->buffer);
        a->buffer = (char *)0;
        a->allocated = a->used = 0;
        free(a);
    }
    return(0);
}

static int aput(a, text)
/*++++++++++++++++
.PURPOSE  Append a string in buffer
.RETURNS  Length of string
-----------------*/
  AREA *a;	/* IN: Area to use  */
  char *text;	/* IN: Text to save */
{
  int len;
    len = 1 + strlen(text);
    while ((len + a->used) > a->allocated) {
	if (a->increment <= 0)	a->increment = a->allocated;
	a->allocated += a->increment;
	a->buffer = realloc(a->buffer, a->allocated);
	if (!a->buffer) error("Not enough memory");
    }
    strcopy(a->buffer + a->used, text);
    a->used += len;
    return(len);
}

static char *asearch(a, text)
/*++++++++++++++++
.PURPOSE  Search a text in the area, case insensitive
.RETURNS  Matched string
.REMARKS  Returned pointer is changed when area is enlarged!!
-----------------*/
  AREA *a;	/* IN: Area to use  */
  char *text;	/* IN: Text to save */
{
  char *p, *e;
    if (a) for (p=a->buffer, e=p+a->used; p<e; p += strlen(p)+1) {
	if (strcasecmp(p, text) == 0)
	    return(p);
    }
    return((char *)0);
}

/*==================================================================
		List of files in directories
 *==================================================================*/

static int find(base, dir, name, area)
/*++++++++++++++++
.PURPOSE  Get directories with specified name
.RETURNS  Number of directories
-----------------*/
  char *base;	/* IN: The base	directory	*/
  char *dir;	/* IN: Sub-directory name 	*/
  char *name;	/* IN: Name of file to find	*/
  AREA *area;	/* MOD: Area with names	   	*/
{
  DIR *dirp; struct dirent *dire;
  struct stat buf;
  char owd[LEN_CWD] , cwd[LEN_CWD] ;
  char *p;
  int ns, i ;

    if (!getcwd(owd, sizeof(owd))) {
	if (!oq) perror("#***getcwd");
    }
    ns = 0;
    if (!dir) dir = "" ;
    p = (*dir ? dir : ".") ;
    dirp = opendir(p);
    if (!dirp) {
	if (!oq) {
	    fprintf(stderr, "#+++opendir %s/./", owd);
	    perror(p);
	}
	return(0);
    }
    if (chdir(p)) {
	if (!oq) {
	    fprintf(stderr, "#+++ch__dir %s/./", owd);
	    perror(p);
	}
	return(0);
    }

    /* Compute the current directory at the required level */
    strcpy(cwd, base);
    if (*cwd) strcat(cwd, "/");
    if (*dir) strcat(cwd, dir);
    if (verbop>1) printf("\t%s", cwd), fflush(stdout);

    while ((dire = readdir(dirp))) {
	if (strcmp(dire->d_name, name) == 0) {	/* A correct directory */
	    aput(area, cwd);
	    ns++;
	    continue;
	}
	if (*(dire->d_name) == '.') continue ;
	i = strloc(dire->d_name, '%');
	if (dire->d_name[i]) continue;
	if (lstat(dire->d_name, &buf)) {	/* Use Link Only */
	    perror(verbop ? "\n***" : "***");
	    fprintf(stderr, "****Can't stat file '%s' in directory '%s/%s'\n",
		dire->d_name, owd, cwd);
	    continue;
	}
	if (S_ISLNK(buf.st_mode)) {
	    if (verbop) fprintf(stderr, "\n");
	    if (!oq) fprintf(stderr, 
		"++++Link '%s' in directory '%s/%s' ignored\n",
		dire->d_name, owd, cwd);
	}
	if (S_ISDIR(buf.st_mode))
		ns += find(cwd, dire->d_name, name, area);
    }
    closedir(dirp);
    ChDir(owd) ;
    return(ns);
}

static AREA *lsdir(dirname)
/*++++++++++++++++
.PURPOSE  Get the subdirectories in a directory
.RETURNS  A newly allocated area with subdirectory names
-----------------*/
  char *dirname; /* IN: Directory name */
{
  DIR *dirp; struct dirent *dire;
  char cwd[LEN_CWD];
  struct stat buf;
  AREA *area;

    dirp = opendir(dirname);
    if (!dirp) {
	/* fprintf(stderr, "#+++opendir '%s': ", dirname);
	   perror(""); */
	return(0);
    }

    if (!getcwd(cwd, sizeof(cwd))) perror("#***getcwd");
    ChDir(dirname);
    area = alloc(BUFSIZ);
    while ((dire = readdir(dirp))) {
	if (dire->d_name[0] == '.') continue ;
	if (strchr(dire->d_name, '%'))	/* Always a link */
	    continue;
	if (lstat(dire->d_name, &buf)) { /* Use Link Only */
	    fprintf(stderr, "#***Can't stat %s/%s\n", dirname, dire->d_name);
	    continue;
	}
	if (S_ISLNK(buf.st_mode)) 	/* Ignore link */
	    continue;
	if (S_ISDIR(buf.st_mode))
	    aput(area, dire->d_name);
    }
    ChDir(cwd);
    closedir(dirp);
    return(area);
}

static AREA *getfiles(base, name)
/*++++++++++++++++
.PURPOSE  Get (recursively) all Files with specified name, keeping only
	directory
.RETURNS  Area filled with names.
-----------------*/
  char *base;	/* IN: The base	directory	*/
  char *name;	/* IN: Name of file to find	*/
{
  char *p;
  int  ns, i;
  AREA *a, *as;
  char **ss, **ps;
  char cwd[LEN_CWD];

    if (!cats) set_catroot();
    if (!getcwd(cwd, sizeof(cwd))) perror("#***getcwd");
    if (verbop) printf("... search for file '%s' from: %s\n", name, cwd);
    a = alloc(BUFSIZ);
    ns = find("", base, name, a);

    /* Sort the catalogues according to Numbers */
    ss = (char **)malloc(ns * sizeof(char*));
    for (i=ns,p=a->buffer,ps = ss; --i>=0; p += 1+strlen(p), ps++) *ps = p;
    qsort(ss, ns, sizeof(char *), (SORT_FCT)filcmp);
    as = alloc(a->used + 1);
    for (i=ns, ps = ss; --i >= 0; ps++) aput(as, *ps);
    afree(a); free(ss);
    if (verbop) printf("\t.\n");
    return(as);
}

/*==================================================================
		Default Catalogues
 *==================================================================*/
static char *current_date(mod)
/*++++++++++++++++
.PURPOSE  Find out the catalogue corresponding to current date
.RETURNS  What this is ...
-----------------*/
    char *mod;	/* IN: Modifier (a, b...) */
{
  static char date[20];
  char *pj, *pd, *ctime(); long t;
    time(&t); pj = ctime(&t); pd = date; if(pj[8] == ' ') pj[8] = '0';
    *(pd++) = 'Y'; *(pd++) = '/';
    *(pd++) = pj[20]; *(pd++) = pj[21]; *(pd++) = pj[22]; *(pd++) = pj[23];
    *(pd++) = '/';
    *(pd++) = pj[4];  *(pd++) = pj[5];  *(pd++) = pj[6];
    *(pd++) = '/';
    *(pd++) = pj[8];  *(pd++) = pj[9];
    if ((strlen(mod) + (pd-date)) < (sizeof(date)-1)) pd += strcopy(pd, mod);
    *pd = 0;
    return(date);
}

static char *current_catalog()
/*++++++++++++++++
.PURPOSE  Find out the catalogue corresponding to current directory
.RETURNS  What this is ...
-----------------*/
{
  char *p;
    p = basename(bwd, 3);	/* Keep 4 levels in directory structure */
    if (strstr(p, "/AFOEV/")) return("AFOEV");
    if (strstr(p, "/afoev/")) return("afoev");
    if ((*p == 'J') && (p[1] == '/'))	/* J-directory */
	return(p);
    if ((*p == '@') && (p[1] == '/'))	/* @-directory */
	return(p);
    if ((*p == 'Y') && (p[1] == '/'))	/* @-directory */
	return(p);
#if 0
    if ((*p == 'c') && (p[1] == 'a') && (p[2] == 't') && (p[3] == 'x')
		  && (p[4] == '/')) { p += 3; *p = '@'; return(p); }
#endif
    return(basename(bwd, 1));	/* Just two levels */
}

/*==================================================================
		Catalogue Name Manipulation
 *==================================================================*/
static int catlen(fname)
/*++++++++++++++++
.PURPOSE  From catalogue designation, find length of directory part
.RETURNS  Length, i.e. of '/' starting the filename
-----------------*/
  char *fname;       /* IN: File catalogue/file */
{
  char *p, **a;
  int c, ns ;

    if ((p = strstr(fname, "/./")))	/* Already defined */
	return(p-fname);

    /* Find the level in the catalogue hierarchy, stored in ns */
    c = toupper(*fname);
    ns = 1;				/* Default: catalog in its directory */
    if (fname[1] == '/') {
        if ((c == 'J') || (c == 'Y') || (c == '@')) ns = 4 ;
	else ns = 2 ;			/* Includes X/... */
    }
    else if (isdigit(*fname)) ns = 1 ;
    else if ((c == 'I') || (c == 'V')) {	/* e.g. III */
	for (p=fname+1; isalpha(*p); p++) {
	    c = toupper(*p);
	    if ((c != 'I') && (c != 'V') && (c != 'X')) break;
	}
	if ((p[0] == '/') && isdigit(p[1]))
	    ns = 2;
    }
    else 	/* Look up among special_names */
    for (a=special_names; *a; a++) {
	if (strncmp(fname, *a+1, strlen(*a+1))) continue;
	ns = *a[0];
	break;
    }
    p = fname ;
    while (*p) {
	if (*p != '/') { p += strloc(p, '/'); continue; }
	if (--ns == 0) break ;
	p++ ;
    }
    return(p-fname) ;
}

static int check_version(dirname)
/*++++++++++++++++
.PURPOSE  Check (by moving through file system) the version (appended A...)
.RETURNS  The version (as letter 'A' , etc...) / -1 for none.
.REMARKS  cd_status is changed
-----------------*/
  char *dirname;       /* IN: String to interpret */
{
  char buf[128], cwd[LEN_CWD] ;
  int len, nv ;
    if (!getcwd(cwd, sizeof(cwd))) perror("#***getcwd");
    cd_status = chdir(dirname) ;
    if (cd_status == 0) {
	chdir(cwd) ;
	return(0) ;
    }

    /* The directory doesn't exist. Try with successive versions */
    len = strlen(dirname) ;
    if (len >= sizeof(buf)-1) {
	fprintf(stderr,
	    "****Directory name > %d bytes: %s\n", 
	    (int)(sizeof(buf)-1), dirname) ;
	exit(1) ;
    }
    strcopy(buf, dirname) ; buf[len+1] = 0 ;
    nv = vtest ? vtest : 'A' ;
    while(1) {
	buf[len] = nv ;
	cd_status = chdir(buf) ;
	if (cd_status == 0) break ;
	if (++nv >= 'I') { nv = -1 ; break; }
    }
    ChDir(cwd) ;
    return(nv) ;
}

static int trN(char *numeric, char *result)
/*++++++++++++++++
.PURPOSE  Convert Numeric Catalogue to e.g. V/NNN, AND separates Catalogue/File
.RETURNS  Converted --- the static catVersion is set to -1 if not existing.
-----------------*/
  /* char *numeric (IN): String to interpret */
  /* char *result (OUT): Result, must be MAX_LINE+6) */
{
  char *p ;
  int nc, cl, ofile, i;
  int stat = 0;	/* Return status */

    while (isspace(*numeric)) numeric++ ;
    ofile = catlen(numeric) ;

    if (isdigit(*numeric)) {		/* Must verify Version.	*/
	p = numeric ;
        cl = *(p++) - '0'; nc = atoi(p);
	if (isADC) sprintf(result, "%c/%d", *numeric, atoi(numeric)) ;
        else sprintf(result, "%s/%d", rm[cl], nc);
	while (isdigit(*p)) p++ ;
	if (isalpha(*p)) {		/* First test version	*/
	    vtest = toupper(*p) ;
	    p++;
	}
	ofile = (p-numeric);
	if ((catVersion = check_version(result))) {
	    p = result + strlen(result) ;
	    if (catVersion != '\377') *(p++) = catVersion ;
	    *p = 0 ;
	}
	vtest = 0 ;
    }
    else if (*numeric) {
	strncpy(result, numeric, ofile) ;  result[ofile] = 0 ;

	/* Convert to uppercase if necessary 	*/
	if (result[1] == '/') {
	    result[0] = toupper(result[0]);
	    if (result[0] == 'B') {		/* B/... always lowercase   */
	        for (p = result+2; *p; p++) *p = tolower(*p);
	    }
	    else if (result[0] != 'X') {	/* X cats is case-dependent */
	        for (p = result; *p; p++) *p = toupper(*p);
	    }
	    /* Special Journal Names */
	    if (starts_by(result, "J/")) {
		if (istarts_by(result, "J/OTHER"))
		    for (i=2; i<7; i++) result[i] = tolower(result[i]);
		else if (starts_by(result, "J/APJ")) result[3] = 'p';
		else if (starts_by(result, "J/AZH")) result[4] = 'h' ;
		else if (starts_by(result, "J/ACA")) result[3] = 'c' ;
		else if (starts_by(result, "J/PAZH")) result[5] = 'h' ;
		else if (starts_by(result, "J/BALT")) {
		    result[3] = 'a'; result[4] = 'l'; result[5] = 't';
		}
	    }
	    else if ((result[0] == 'Y')||(result[0] == '@')) {   /* Date */
		/* Form Y/yyyy/Mon/day */
		if (strlen(result)>7) for (i=8; isalpha(result[i]); i++) 
		    result[i] = tolower(result[i]);
	    }
	}
    }
    /* Added V3.7: add 'A' for A&A (from vol.509, articles are numbered) */
    if ((starts_by(result, "J/A+A/")) && isdigit(result[6])) {
      int v = 0; char *t;
	if ((t = strchr(result+6, '/')) && isalnum(t[1]))
	    v = atoi(result+6);
	if (v >= 509) {
	    t++;
	    if (isdigit(*t)) {      /* J/A+A/509/94 --> J/A+A/509/A94 */
	        p = t + strlen(t); p[1] = 0;
	        while (--p>=t) p[1] = p[0];
	        *t = 'A';
	    }
	    else if (islower(*t))  /* J/A+A/509/a94 --> J/A+A/509/A94 */
		*t = toupper(*t);
	}
    }

    /* Verify the 'other' directories */
    if (starts_by(result, "J/other/")) { AREA *area; char x, *s;
	area = lsdir("J/other");
	s = result+8;
	nc = strloc(s, '/');
	x = s[nc]; s[nc] = 0;
	if ((p = asearch(area, s))) {	/* Found the name, case could differ */
	    for (i=0; p[i]; i++) s[i] = p[i];
	    s[nc] = x;
	}
	else {	/* Could be e.g. J/other/BaltA/13.1 --> J/BaltA/13/1 */
	    afree(area);
	    area = lsdir("J");
	    if ((p = asearch(area, s))) {	/* Yes, J/other/jnl => J/jnl */
		for (i=0; p[i]; i++) result[2+i] = p[i];
		result[2+i] = 0;	/* Terminate string */
		/* Mod. V3.81: from e.g. J/other/ApJ/135/12 */
		if (x) { int idot = 0;
		     s[nc] = x; p = s+nc;	/* p-> what remains to copy */
		     s = result +2+i;		/* s-> continuation */
		     for (i=0; p[i]; i++) {
			 s[i] = p[i];
			 if (s[i] == '.') idot = i;
		     }
		     s[i] = 0;		/* Terminate string */
		     if (idot) s[idot] = '/';	/* e.g. 13.1 => 13/1 */
		     else if (numeric[ofile] == '/') {
			 /* e.g. J/other/ApJ/135/12: one level more */
			 s[i++] = numeric[ofile++];
			 while (numeric[ofile] && (numeric[ofile]!='/'))
			     s[i++] = numeric[ofile++];
		         s[i] = 0;	/* Terminate string */
		     }
		}
	    }
	    else s[nc] = x;
	}
	afree(area);
    }

    /* Add the file now */
    if (numeric[ofile]) {
	p = numeric + ofile ;
	if (p[0] != '/') strcat(result, p);
	else if (optf) {	/* Insert /./ */
	    if ((p[0] == '/') && (p[1] == '.') && (p[2] == '/'))
	        p += 2 ;	/* /./ already there */
	    if (p[1]) {	/* Not just a / */
	       strcat(result, "/.");
	       /* convert lowercase readme to ReadMe */
	       if (strcasecmp(p, "/readme") == 0) p = "/ReadMe";
	       strcat(result, p);
	    }
	}
	else strcat(result, p);
    }

#if 0
    /* Convert Blank in file name to + (V3.26) */
    if ((p = strchr(result, ' ')))
	while(*p) { if (isspace(*p)) *p = '+'; p++; }
#else	/* V3.6: convert blanks to underscores */
    if ((p = strchr(result, '\37')))
	while(*p) { if (iscntrl(*p)) *p = '_'; p++; }
#endif
    return(stat);
}

static int trn(char *alpha, char *result)
/*++++++++++++++++
.PURPOSE  Convert Alphabetic Catalogue to Numeric
.RETURNS  0=OK + result
-----------------*/
  /* char *alpha    (IN): String to interpret */
  /* char *aresult (OUT): Result */
{
  char *p, *pc;
  int i, nc ;
  int stat = 0; 		/* Returned status */

    for (p = alpha; isspace(*p); p++) ;	alpha = p;
    if (isdigit(*p)) {
	strcpy(result, alpha);
	return(stat);
    }
    for (i=0; isupper(*p); p++)	result[i++] = *p; result[i] = 0;
    for (i=0; i < ITEMS(rm); i++) {
	if (strcmp(rm[i], result) == 0) break;
    }
#if 0
    if (i >= ITEMS(rm))	{	/* Take care of HTML Syndrome */
	while (*p) {
	    if ((p[0] == ' ') && isalnum(p[1])) *(p++) = '+' ;
	    p++ ;
	}
	strcpy(result, alpha);
	return(stat);
    }
#endif
    if (*p == '/') p++;
    if (i<ITEMS(rm)) {
  	nc = atoi(p);	while(isdigit(*p)) p++;
	sprintf(result, "%d%s", i*1000+nc, p);
    }
    else sprintf(result, "0%s", p);

    pc = result + strlen(result) - 1;
    if (*pc == '/')	*pc = 0;

#if 0
    /* Convert Blank in file name to + (V3.26) */
    if ((p = strchr(result, ' ')))
	while(*p) { if (isspace(*p)) *p = '+'; p++; }
#else	/* V3.6: convert blanks to underscores */
    if ((p = strchr(result, '\37')))
	while(*p) { if (iscntrl(*p)) *p = '_'; p++; }
#endif
    return(stat);
}

static int standardize(char *text)
/*++++++++++++++++
.PURPOSE  Standardize a string representing a catalog (HTML syndrome)
.RETURNS  0 = not modified, 1=modified 2=contains spaces(replaced by US) 3=both
-----------------*/
{
  char *p, *t;
  int i, x, st=0;

    /* %xx: replace by char */
    t = text;
    while ((t = strchr(t, '%'))) {
	if (isxdigit(t[1]) && isxdigit(t[2])) {
	    x = isalpha(t[1]) ? t[1]+9 : t[1]; i  = (x&0xf)<<4;
	    x = isalpha(t[2]) ? t[2]+9 : t[2]; i |=  x&0xf;
	    t[0] = i;
	    for (p=t+2, i=1; p[i]; i++) t[i] = p[i];
	    t[i] = 0;    /* Double encoding is resolved */
	    st = 1;
	}
	else t++;
    }

    /* Remove the leading/trailing blanks */
    for (t=text; *t == ' '; t++) ;
    if (t != text) {
	for (p=text; *t; p++, t++) *p = *t;
	st = 1;
	*p = 0;
    }
    t = text+strlen(text)-1;
    if (*t == ' ') {
	while ((t>text) && (*t == ' ') ) *(t--) = 0;
	st = 1;
    }

    /* Other blanks: replace by US (31=\37) */
    t = text;
    while ((t = strchr(t, ' '))) { *t = '\37'; st |= 2; }

    /* Ampersand: replace by '+' if not HTML */
    t = text;
    while ((t = strchr(t, '&'))) {
	for (i=0; htmspec[i]; i++) {
	    if (strncmp(t, htmspec[i]+1, strlen(htmspec[i]+1)) == 0)
		break;
	}
	if (htmspec[i]) {
	    p = t + strlen(htmspec[i]+1);
	    *(t++) = htmspec[i][0];
	    while (p[i]) { t[i] = p[i]; i++; }
	    *p = 0;
	}
	else *(t++) = '+';
	st = 1;
    }

    /* Accept J_jnl as J/jnl */
    if (isalpha(*text) && (t = strchr(text, '_'))) {
	for (p = text; p<t; p++) {
	    if (*p == '/') break;
	    x = toupper(*p);
	    if (!strchr("IJVXB", x)) break;
	}
	if (p == t) { *t = '/'; st = 1; }
    }
    /* Replace US (31=\37) for journals with '+' */
    if ((st&2) && (text[0] == 'J') && (text[1] == '/')) {
	for (i=2; text[i] && (text[i] != '/'); i++) {
	    if (text[i] == '\37') text[i] = '+';
	}
	/* No more blank ? */
	if (strchr(text+i, '\37')) ; else st &= ~2;
    }
    /* Missing J ? */
    if (strncasecmp(text, "A+A", 3) == 0) {
	for (t = text+strlen(text); t>=text; t--) t[2] = t[0];
	text[0] = 'J';
	text[1] = '/';
	st = 1;
    }
    return(st);
}

static int trd(char *catid, char *catname)
/*++++++++++++++++
.PURPOSE  Move to directory defined by catid if it exists.
.RETURNS  0=OK / non-zero=err
-----------------*/
  /* char *catid    (IN): String to interpret */
  /* char *catname (OUT): Actual catalog name, must be MAX_LINE+20 */
{
  int cd_status = 128;
  int len, c;
  char *p;

    ChDir(cats) ;
    if (isdigit(catid[0]) && isdigit(catid[1])
     && isdigit(catid[2]) && isdigit(catid[3])) {	/* Just Numeric */
	if (isADC) { /* ADC-Style */
	    strcpy(catname, "catalogs") ;
	    if (chdir(catname)){
		fprintf(stderr,
		    "****Can't move to '%s' for '%s'\n",
		    catname, catid) ;
		return(cd_status) ;
	    }
	    /* catdir = trN(catid, catdir) ; */
	    len = strlen(catname);
	    catname[len++] = '/';
	    trN(catid, catname+len);
	}
	else {	/* At CDS */
	    trN(catid, catname);
	}
	goto TestDir ;
    }
    /* Not a numeric number. Therefore if starts by I, V or X,
       assume a numbered catalog.
    */
    if ((toupper(catid[0]) == 'I') || (toupper(catid[0]) == 'V')
      ||(toupper(catid[0]) == 'X')) { char catnum[MAX_LINE+6];
	trn(catid, catnum);
	if (isdigit(catnum[0])) return(trd(catnum, catname)) ;
    }

    /* A journal table starts by 'J' */
    if (isADC && (toupper(catid[0]) == 'J') && (catid[1] == '/')) {
	strcpy(catname, "journal_tables") ;
	if (chdir(catname)){
	    fprintf(stderr,
		"****Can't move to '%s' for '%s'\n",
		catname, catid) ;
	    return(cd_status) ;
	}
	len = strlen(catname)-1; c = catname[len];
	trN(catid, catname+len);	/* ADC ignores the "J/" */
	catname[len++] = c;  		/* replace 'J' */
	cd_status = 128 ;			/* Force Verification   */
	goto TestDir ;
    }

    /* Other possibilities: just try... */
    /* catdir = trN(catid) ; strcpy(catname, catdir) ; */
    trN(catid, catname);

    /* Test if the directory effectively exists */
  TestDir:
    if (cd_status&128) {
        getcwd(awd, sizeof(awd));
	if ((p = strstr(catname, "/./"))) *p = 0 ;
	if (verbop) fprintf(stderr, "#...TestDir %s => %s", awd, catname);
	if (*catname) cd_status = chdir(catname) ;
	else          cd_status = 0;
	if(cd_status && verbop) {
	    perror(" ");
	}
	if (verbop) fprintf(stderr, "\n");
	if (p) *p = '/' ;
    }
    return(cd_status);
}

/*==================================================================*/
static int adelete(a1, not)
/*++++++++++++++++
.PURPOSE  Delete in area a1 the list in not
.RETURNS  Number of deleted elements
-----------------*/
  AREA *a1;	/* Area with filenames */
  AREA *not;	/* Area with excluded filenames */
{
  char *p1, *p2, *p;
  int n, n1, n2, l1, l2;

    if (not) n2 = not->used, p2 = not->buffer;
    else	   n2 = 0, p2 = (char *)0;
    n = 0;

    for (n1 = a1->used, p = p1 = a1->buffer; n1 > 1; n1 -= l1, p1 += l1) {
	l1 = 1 + strlen(p1);
	while ( (n2 > 0) && (filcmp(&p1, &p2) > 0)) {
		l2 = 1 + strlen(p2);
		p2 += l2; n2 -= l2;
	}
	if ((n2 > 0) && (filcmp(&p1, &p2) == 0)) {
		l2 = 1 + strlen(p2);
		p2 += l2;
		n2 -= l2;
		n++;
		continue;
	}
	if (n) memcpy(p, p1, l1);
	p += l1;
    }
    return(n);
}

static int edx(a1, not)
/*++++++++++++++++
.PURPOSE  Edit strings in a1, but not in a2
.RETURNS  Number of printed lines
-----------------*/
  AREA *a1;	/* Area with filenames */
  AREA *not;	/* Area with excluded filenames */
{
  char *p1, *p2;
  int n, n1, n2, l1, l2;

    if (not) n2 = not->used, p2 = not->buffer;
    else	   n2 = 0, p2 = (char *)0;
    n = 0;

    for (n1 = a1->used, p1 = a1->buffer; n1 > 1; n1 -= l1, p1 += l1) {
	l1 = 1 + strlen(p1);
	while ( (n2 > 0) && (filcmp(&p1, &p2) > 0)) {
		l2 = 1 + strlen(p2);
		p2 += l2; n2 -= l2;
	}
	if ((n2 > 0) && (filcmp(&p1, &p2) == 0)) {
		l2 = 1 + strlen(p2);
		p2 += l2;
		n2 -= l2;
		continue;
	}
	puts(p1); n++;
    }
    return(n);
}

/*==================================================================
		Program
 *==================================================================*/
int main (int argc, char **argv)
{
  int nc, len, i, argi, argu, status=0;
  static int (*af)() = (int(*)()) 0;
  int (*fct)();
  char line[MAX_LINE], *p, *d, *catroot, result[MAX_LINE+20] ;
  char *comment=(char *)0, comment_char;
  AREA *a, *obsolete, *private;

    nc  = 0;

	/* Check the Arguments */

    if (!getcwd(bwd, sizeof(bwd))) perror("#***getcwd");
    pgm = basename(*argv, 0);

    catroot = "" ;

    for (argi=1; argi<argc; argi=argu) { /*while (-- argc > 0) */
	p = argv[argi]; argu = argi+1; /*p = *++argv;*/
	a = obsolete = private = (AREA *)0;
	if (*p == '-') for (++p; *p; p++) switch(*p) /*switch (*++p)*/ {
	  case 'c' :	/* May be -cat= */
	    if (starts_by(p, "cat=")) break ;
	    goto bad_option;
	  case 'd' :
	    af = trd;  	/* Convert to Directory Name */
	    if (p[1] == 'f') optf = *++p;	/* Letter of option */
	    /* optf = p[1] ;*/
	    if (!cats) set_catroot();
	    catroot = malloc(strlen(cats)+2) ;
	    strcpy(catroot, cats) ;
	    strcat(catroot, "/") ;
	    continue;
	  case 'r' :		/* Relative Path to Catalog  */
	    af = trd;
	    if (p[1] == 'f') optf = *++p;	/* Letter of option */
	    catroot = "" ;
	    continue ;
	  case '1' :            /* Just 1 catalog/line	*/
	    o1 |= 1 ;
	    continue ;
	  case 'n' :
	    af = trn;  	/* Convert to Numeric */
	    if (strchr(p, '1')) o1 |= 1 ;
	    continue;
	  case 'N' :
	    af = trN;  	/* Convert to Roman   */
	    if (strchr(p, '1')) o1 |= 1 ;
	    continue;
	  case 'q':	/* Quiet */
	    oq = 1;
	    continue;
	  case 'v':                     /* Verbose      */
	    if (starts_by(p, "ver")) {
	        printf("trcat (CDS) Version %s\n", VERSION);
	        exit(0);
	    }
	    if (isdigit(p[1])) {	/* Level of verbosity */
		verbop = atoi(p+1);
		while (isdigit(p[1])) p++;
	    }
	    else verbop = 1;
	    /* fprintf(stderr, "... %s: Examine argument ", pgm+4); */
	    continue;
	  case 'o':			/* Get Obsolete Files */
	    if (argu>=argc) error("Missing argument after -", p);
	    obsolete = getfiles(argv[argu++], "=obsolete=");
	    /* if (--argc <= 0) p = ""; else p = *++argv ;
	    obsolete = getfiles(p, "=obsolete="); */
	    nc++;
	    edx(obsolete, (AREA *)0);
	    continue;
	  case 'p':			/* Get Private Files */
	    if (argu>=argc) error("Missing argument after -", p);
	    private = getfiles(argv[argu++], "=PRIVATE=");
	    nc++;
	    edx(private, (AREA *)0);
	    continue;
	  case 'x':			/* Remove Old	*/
	    if (argu>=argc) error("Missing argument after -", p);
	    obsolete = getfiles(argv[argu], "=obsolete=");
	    /*if (--argc <= 0) error("Missing argument after -", p);
	    obsolete = getfiles(*++argv, "=obsolete="); */
	    goto optX;
	  case 'X':
	    if (argu>=argc) error("Missing argument after -", p);
	    /*if (--argc <= 0) error("Missing argument after -", p);
	    ++argv ; */
	  optX:
	    private = getfiles(argv[argu], "=PRIVATE=");
	    a = getfiles(argv[argu++], ".files");
	    if (private) {
		if (private->used) adelete(a, private);
		afree(private);
		private = 0;
	    }
	    /*a = getfiles(*argv, ".files"); */
	    nc++;
	    edx(a, obsolete);
	    afree(a); if (obsolete) afree(obsolete);
	    continue;
	  case 'h':                     /* Help         */
	    help();
	    continue;
	  default:
	  bad_option:
	    error("Bad option: ", p);
	    p += strlen(p)-1;
	    continue;
	}
	if (!*p) continue;

	/* Not an option here -- we have the name of a catalog */
	/* Skip the "cat=" or "-cat=" */
	if (strlen(p) >= MAX_LINE)
	   error("Too long expression: ", p);
	if (!cats) set_catroot();
	if (starts_by(p, "cat=")) p += 4;
	fct = af;
	if (!fct) fct = isdigit(*p) ? trN : trn;
	switch(*p) {
	  case '.':
	    p = current_catalog();
	    break;
	  case 'Y':
	  case '@':
	    switch(p[1]) {
	      case '+': case '-':	/* Relative to current date */
		d = current_date("");
		sprintf(d+11, "%d", atoi(d+11) + atoi(p+1));
		p = d;
		break;
	      case '/':		/* Check Y/dd */
		d = current_date("");
		if (isdigit(p[2]) && isdigit(p[3]) && (!p[4]))
		    strcpy(d+11, p+2), p = d;
		break;
	      default:
		p = current_date(p+1);
	    }
	    break;
	  case 'a':
	    if (starts_by(p, "afoev/"))  break; /* p = "afoev"; */
	  case 'A':
	    if (starts_by(p, "AFOEV/"))  break; /* p = "afoev"; */
	  case 'J': case 'F': case 'M': case 'S': case 'O':
	  case 'N': case 'D':
	  case 'j': case 'f': case 'm': case 's': case 'o':
	  case 'n': case 'd':
	    if (starts_by(p, "afoev/")) break;	/* OK */
	    if (starts_by(p, "AFOEV/")) break;	/* OK */
	    /* Month / day ; we assume an existing Directory */
	    if (islower(p[1]) && islower(p[2]) && (p[3]=='/') && isdigit(p[4])
	      && (strlen(p)<=6)) {
		d = current_date("");
		strcpy(d+7,p); d[7] = toupper(d[7]);
		if (isdigit(d[11]) && (!d[12])) {
		    d[12] = d[11]; d[11] = '0';
		    d[13] = 0;
		}
		p = d;
		if (chdir(p)) {	/* No such directory ! */
		    i = atoi(p+2) - 1 ;
		    sprintf(p+2,"%d", i);
		    p[6] = '/';
		}
		ChDir(cats);
	    }
	    break;
	  default:
	    if (isdigit(*p) && (p[4] == '/') && isupper(p[5])
		&& islower(p[6]) && islower(p[7]) && (p[8] == '/')
		&& isdigit(p[9]) && isdigit(p[10]) && (!p[11])) {
		d = current_date("");
		strcpy(d+2, p);
		p = d;
		break;
	    }
	}
	/* Standardize: remove effects of HTML syndrome... */
	strncpy(line, p, sizeof(line)-2); line[sizeof(line)-2] = 0;
	if (standardize(line) && verbop) printf(
		"... converted catalog into '%s'\n", line);
	status |= (*fct)(line, result) ;
	if (*result && (status==0)) printf("%s%s\n", catroot, result) ;
	else ; /* printf("++++Catalog:%s\n", p) ; */
	nc++;
	continue;
    }
    if (nc)    exit(status ? 1 : 0);
    if (!cats) set_catroot();

    while(fgets(line, sizeof(line)-2, stdin)) {
	/* Remove trailing blanks */
	p = line + strlen(line) ;
	while (--p >= line) { if (isgraph(*p)) break ; }
	if (++p <= line) continue ;	/* BLANK line ! */
	*p = 0 ;
	if (standardize(line) && verbop) printf(
		"... converted catalog into '%s'\n", line);

	/* Keep the Comment */
	if (o1 && isgraph(line[0])) {
	    for (p=line; isgraph(*p); p++) ;
	    comment_char = *p ;	/* comment_char = 0 when no comment */
	    *p = 0;
	    comment = p+1;
	}
	else comment_char = 0 ;

	/* V3.0: Remove the directory path if starts by / */
	p = line;
	if (line[0] == '/') {
	    len = strlen(CATroot[iCATroot]);
	    if (strncmp(line, CATroot[iCATroot], len) == 0)
		p = line + len;
	    else for (i=0; i<ITEMS(CATroot); i++) {
		len = strlen(CATroot[i]);
		if (strncmp(line, CATroot[i], len)) continue;
		p = line + len;
		break;
	    }
	}

	/* Which function should we apply ? */
	fct = af;
	if (!fct)  fct = isdigit(*p) ? trN : trn;
	status |= (*fct)(p, result) ;
	/* if (*result) { */
	    printf("%s%s", catroot, result) ;
	    if (comment_char) printf("%c%s", comment_char, comment) ;
	    putchar('\n') ;
	/* }
	else printf("++++Catalog:%s\n", line) ;
	 */
    }

    exit(status ? 1 : 0);

}
